import time
from typing import Type

import pytest

from squeal import Queue
from .common import TemporaryMySQLBackend, TemporaryLocalBackend, TemporaryBackendMixin


@pytest.mark.parametrize(
    "backend_class", [TemporaryMySQLBackend, TemporaryLocalBackend]
)
class TestMySQLQueue:
    def test_queue_topics_dont_interfere(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            q = Queue(bk)

            q.put(b"a", topic=1)
            assert q.get(topic=2) is None

            q.put(b"b", topic=2)
            assert q.get(topic=2) is not None

    def test_queue_topics(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            for _ in range(1):
                q.put(b"", topic=1)
            for _ in range(5):
                q.put(b"", topic=2)
            for _ in range(4):
                q.put(b"", topic=3)

            topics = dict(q.list_topics())
            assert {1: 1, 2: 5, 3: 4} == topics

            assert 0 == q.get_topic_size(100)
            assert 1 == q.get_topic_size(1)
            assert 5 == q.get_topic_size(2)
            assert 4 == q.get_topic_size(3)

            x = q.get(topic=3)
            assert x is not None

            assert 3 == q.get_topic_size(3)

    def test_queue_put_get_destroy(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.put(b"test_queue_put_get_destroy", topic=1)
            ret = q.get(topic=1)

            assert b"test_queue_put_get_destroy" == ret.payload

    def test_get_nothing(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            assert q.get(topic=1) is None

    def test_no_double_get(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)

            q.put(b"test_no_double_get", topic=1)
            ret = q.get(topic=1)

            assert b"test_no_double_get" == ret.payload

            assert q.get(topic=1) is None

    def test_queue_automatic_release(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk, visibility_timeout=1)
            q.put(b"test_queue_automatic_release", topic=1)

            x = q.get(topic=1)
            assert b"test_queue_automatic_release" == x.payload

            time.sleep(2)

            y = q.get(topic=1)
            assert b"test_queue_automatic_release" == y.payload

    def test_queue_nack(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk, failure_base_delay=0)
            q.put(b"test_queue_nack", topic=1)

            x = q.get(topic=1)
            assert b"test_queue_nack" == x.payload

            assert q.get(topic=1) is None

            x.nack()
            z = q.get(topic=1)
            assert b"test_queue_nack" == z.payload

    def test_queue_skip_nack_with_delay(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            q = Queue(bk, failure_base_delay=100)
            q.put(b"a", topic=1)
            q.put(b"b", topic=1)
            q.put(b"c", topic=1)

            x1 = q.get(topic=1)
            assert b"a" == x1.payload
            x1.nack()

            x2 = q.get(topic=1)
            assert b"b" == x2.payload
            x2.nack()

            x3 = q.get(topic=1)
            assert b"c" == x3.payload
            x3.nack()

            x4 = q.get(topic=1)
            assert x4 is None

    def test_hash_uniqueness(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            assert 1 == q.put(b"", topic=1, priority=0, hsh=b"0000000000000000")
            assert 1 == q.put(b"", topic=1, priority=100, hsh=b"0000000000000001")
            assert 0 == q.put(b"", topic=1, hsh=b"0000000000000001")

            x = q.get(topic=1)
            x.ack()

            assert 1 == q.put(b"", topic=1, hsh=b"0000000000000001")

    def test_batch_put_uniqueness_violation(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            q = Queue(bk)
            assert 1 == q.batch_put([(b"b", 1, b"0000000000000000")], priority=0)
            assert 0 == q.batch_put([(b"b", 1, b"0000000000000000")], priority=0)

    def test_put_with_rate_limit(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk, rate_limit_per_hour=1)
            assert 1 == q.put(item=b"a", topic=1, hsh=b"0000000000000000")
            assert 0 == q.put(item=b"a", topic=1, hsh=b"0000000000000000")
            m = q.get(topic=1)
            m.ack()
            assert 0 == q.put(item=b"a", topic=1, hsh=b"0000000000000000")

    def test_batch_put(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.batch_put([(b"a", 1, None), (b"b", 1, None), (b"c", 1, None)])
            assert 3 == q.get_topic_size(topic=1)
            msgs = q.batch_get(topics=[(1, 3)])
            assert 3 == len(msgs)
            assert 0 == q.get_topic_size(topic=1)

    def test_batch_get(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.put(b"a", topic=1, priority=0)
            q.put(b"b", topic=1, priority=1)

            msgs = q.batch_get(topics=[(1, 2)])
            assert 2 == len(msgs)

    def test_put_get(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.put(b"a", topic=1)
            msg = q.get(topic=1)
            assert b"a" == msg.payload

    def test_delay(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk, new_message_delay=1)
            q.put(b"a", topic=1)
            assert q.get(topic=1) is None
            time.sleep(1.1)
            x = q.get(topic=1)
            assert b"a" == x.payload

    def test_visibility_timeout(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk, visibility_timeout=1)
            q.put(b"a", topic=1)
            q.get(topic=1)
            assert q.get(topic=1) is None
            time.sleep(2)
            x = q.get(topic=1)
            assert b"a" == x.payload

    def test_priority(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.put(b"a", topic=1, priority=0)
            q.put(b"b", topic=1, priority=1)

            msg = q.get(topic=1)
            assert b"b" == msg.payload

    def test_topic_lock_with_no_topics(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            q = Queue(bk)

            topic = q.acquire_topic()
            assert topic is None

    def test_topic_lock(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            q = Queue(bk)
            q.put(b"", topic=1)

            topic = q.acquire_topic()
            assert topic.idx == 1

            topic2 = q.acquire_topic()
            assert topic2 is None

            assert len(q.list_held_topics()) == 1
            topic.release()
            assert len(q.list_held_topics()) == 0

            topic3 = q.acquire_topic()
            assert topic3.idx == 1



import dataclasses
import json
import logging
from typing import Dict, Any, Union

LOGGING_MODE = "normal"


def enable_structured_logging():
    global LOGGING_MODE
    LOGGING_MODE = "structured"


def set_log_level(level: Union[int, str]):
    get_logger().setLevel(level)


def get_logger():
    return logging.getLogger("squeal")


@dataclasses.dataclass
class LogMessage:
    message: str
    fields: Dict[str, Any]

    def __str__(self) -> str:
        if LOGGING_MODE == "normal":
            return f"{self.message} >>> {json.dumps(self.fields)}"
        if LOGGING_MODE == "structured":
            return json.dumps({"message": self.message, **self.fields})


lm = LogMessage



import time
from collections import Counter
from typing import List, Tuple, Iterable, Optional, Collection

from .base import Backend, Message, TopicLock


class LocalBackend(Backend):
    """
    This is just a reference implementation for testing purposes.  Don't use it for anything else.  It's deliberately
    unoptimized to make it easy to follow what it's doing.
    """

    def __init__(self):
        super().__init__()
        self.messages = []
        self.topic_locks = {}
        self.rate_limits = {}
        self.unique_constraint = set()
        self.next_id = 0
        self.created = False
        self._max_payload_size = None
        self._hash_size = 16

    @property
    def max_payload_size(self) -> Optional[int]:
        return self._max_payload_size

    @property
    def hash_size(self) -> int:
        return self._hash_size

    def create(self) -> None:
        self.created = True

    def destroy(self) -> None:
        self.created = False

    def batch_put(
        self,
        data: Collection[Tuple[bytes, int, Optional[bytes]]],
        priority: int,
        delay: int,
        failure_base_delay: int,
        visibility_timeout: int,
    ) -> int:
        assert self.created
        for payload, topic, hsh in data:
            if hsh is not None and len(hsh) != self.hash_size:
                raise ValueError(
                    f"hsh size is not HASH_SIZE ({len(hsh)} != {self.hash_size})"
                )

        constraint_violations = []
        for payload, topic, hsh in data:
            if hsh is not None:
                if (topic, hsh) in self.unique_constraint:
                    constraint_violations.append((payload, topic, hsh))
                    continue
                self.unique_constraint.add((topic, hsh))

            self.messages.append(
                {
                    "id": self.next_id,
                    "payload": payload,
                    "topic": topic,
                    "hsh": hsh,
                    "priority": priority,
                    "acquired": False,
                    "visibility_timeout": visibility_timeout,
                    "delivery_time": time.time() + delay,
                    "failure_base_delay": failure_base_delay,
                    "failure_count": 0,
                }
            )
            self.next_id += 1

        return len(data) - len(constraint_violations)

    def release_stalled_messages(self) -> int:
        assert self.created
        now = time.time()
        n = 0
        for msg in self.messages:
            if not msg["acquired"]:
                continue
            if msg["visibility_timeout"] + msg["acquire_time"] > now:
                continue

            msg["acquired"] = False
            n += 1
        return n

    def batch_get(self, topic: int, size: int) -> List["Message"]:
        assert self.created
        now = time.time()
        self.messages.sort(key=lambda x: [-x["priority"], x["id"]])

        output = []
        for msg in self.messages:
            if len(output) >= size:
                break
            if msg["acquired"]:
                continue
            if msg["topic"] != topic:
                continue
            if msg["delivery_time"] > now:
                continue

            msg["acquired"] = True
            msg["acquire_time"] = now
            output.append(Message(msg["payload"], msg["id"], self))
        return output

    def ack(self, task_id: int) -> None:
        assert self.created
        for idx, msg in enumerate(self.messages):
            if msg["id"] != task_id:
                continue
            if not msg["acquired"]:
                continue
            break
        else:
            return
        if self.messages[idx]["hsh"] is not None:
            k = (self.messages[idx]["topic"], self.messages[idx]["hsh"])
            self.unique_constraint.remove(k)
        del self.messages[idx]

    def batch_nack(self, task_ids: Collection[int]) -> None:
        assert self.created
        to_nack = set(task_ids)
        for idx, msg in enumerate(self.messages):
            if msg["id"] not in to_nack:
                continue
            if not msg["acquired"]:
                continue

            msg["acquired"] = False
            delay = msg["failure_base_delay"] * (2 ** msg["failure_count"])
            msg["failure_count"] += 1
            msg["delivery_time"] = time.time() + delay
            print(delay)

    def batch_touch(self, task_ids: Collection[int]) -> None:
        assert self.created
        to_touch = set(task_ids)
        for msg in self.messages:
            if msg["id"] in to_touch:
                continue
            if not msg["acquired"]:
                continue
            msg["acquire_time"] = time.time()

    def list_topics(self) -> List[Tuple[int, int]]:
        assert self.created
        counts = Counter()

        now = time.time()
        for msg in self.messages:
            if msg["acquired"]:
                continue
            if msg["delivery_time"] > now:
                continue

            counts[msg["topic"]] += 1

        return list(counts.items())

    def get_topic_size(self, topic: int) -> int:
        assert self.created
        n = 0
        now = time.time()
        for msg in self.messages:
            if msg["acquired"]:
                continue
            if msg["topic"] != topic:
                continue
            if msg["delivery_time"] > now:
                continue

            n += 1
        return n

    def acquire_topic(
        self, topic_lock_visibility_timeout: int
    ) -> Optional["TopicLock"]:
        assert self.created
        for topic, size in self.list_topics():
            if topic not in self.topic_locks:
                self.topic_locks[topic] = {
                    "acquire_time": time.time(),
                    "visibility_timeout": topic_lock_visibility_timeout,
                }
                return TopicLock(topic, self)
        return None

    def batch_release_topic(self, topics: Collection[int]) -> None:
        assert self.created
        for topic in topics:
            if topic in self.topic_locks:
                del self.topic_locks[topic]

    def batch_touch_topic(self, topics: Collection[int]) -> None:
        assert self.created
        for topic in topics:
            if topic not in self.topic_locks:
                continue
            self.topic_locks[topic]["acquire_time"] = time.time()

    def release_stalled_topic_locks(self) -> int:
        assert self.created
        to_release = [
            topic
            for topic, meta in self.topic_locks.items()
            if meta["acquire_time"] + meta["visibility_timeout"] < time.time()
        ]
        for topic in to_release:
            del self.topic_locks[topic]
        return len(to_release)

    def rate_limit(self, key: bytes, max_events: int, interval_seconds: int) -> bool:
        if len(key) != self.hash_size:
            raise ValueError(
                f"rate limit key size is not HASH_SIZE ({len(key)} != {self.hash_size})"
            )

        if key not in self.rate_limits:
            self.rate_limits[key] = []

        now = time.time()
        self.rate_limits[key] = [t for t in self.rate_limits[key] if t > now]

        if len(self.rate_limits[key]) < max_events:
            self.rate_limits[key].append(now + interval_seconds)
            return True

        return False



import random
from typing import List, Tuple, Optional, Collection

from .base import Backend, Message, TopicLock

# Create a table to store queue items
# format args:
#   name -> table name
#   hash_size -> hsh size (bytes)
#   size -> max message size (bytes)
SQL_CREATE = """
CREATE TABLE IF NOT EXISTS {name} (
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    topic INT UNSIGNED NOT NULL,
    hash BINARY({hash_size}) NULL,
    priority INT UNSIGNED NOT NULL,
    owner_id INT UNSIGNED NULL,
    delivery_time DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP,
    visibility_timeout INT UNSIGNED NOT NULL,
    failure_base_delay INT UNSIGNED NOT NULL,
    failure_count INT UNSIGNED DEFAULT 0,
    acquire_time DATETIME NULL DEFAULT NULL ON UPDATE CURRENT_TIMESTAMP,
    payload VARBINARY({size}),
    PRIMARY KEY (id),
    UNIQUE (topic, hash)
)
"""
SQL_CREATE_LOCKS = """
CREATE TABLE IF NOT EXISTS {name} (
    topic INT UNSIGNED NOT NULL,
    owner_id INT UNSIGNED NOT NULL,
    visibility_timeout INT UNSIGNED NOT NULL,
    acquire_time DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP,
    PRIMARY KEY (topic)
)
"""
SQL_CREATE_RATE_LIMITS = """
CREATE TABLE IF NOT EXISTS {name} (
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    hash BINARY({key_size}) NULL,
    expire_time DATETIME NOT NULL,
    PRIMARY KEY (id),
    INDEX (hash)
)
"""

# Destroy the table
# format args:
#   name -> table name
SQL_DROP = "DROP TABLE IF EXISTS {name}"

# Insert a message into the queue
# format args:
#   name -> table name
# sql substitution args:
# * payload
# * topic
# * hsh
# * priority
# * delay (seconds)
# * failure_base_delay (seconds)
# * visibility timeout (seconds)
SQL_INSERT = (
    "INSERT INTO {name} (payload, topic, hash, priority, delivery_time, failure_base_delay, visibility_timeout)"
    "VALUES (%s, %s, %s, %s, TIMESTAMPADD(SECOND, %s, CURRENT_TIMESTAMP), %s, %s)"
)

# Release stalled messages
# format args:
#   name -> table name
SQL_UPDATE = """
UPDATE {name} SET owner_id=NULL
    WHERE owner_id IS NOT NULL
    AND TIMESTAMPDIFF(SECOND, acquire_time, CURRENT_TIMESTAMP) > visibility_timeout
"""

# Can't guarantee that columns are updated in a particular
# order in one update statement, so we do this in two steps.
SQL_BATCH_NACK_1 = """
UPDATE {name}
   SET delivery_time=TIMESTAMPADD(SECOND, failure_base_delay * POW(2, failure_count), CURRENT_TIMESTAMP)
   WHERE owner_id=%s AND id IN %s
"""
SQL_BATCH_NACK_2 = """
UPDATE {name}
   SET owner_id=NULL,
       failure_count = failure_count + 1
   WHERE owner_id=%s AND id IN %s
"""

# Refresh the acquire time for a bunch of messages
SQL_BATCH_TOUCH = """
UPDATE {name}
   SET acquire_time = CURRENT_TIMESTAMP
   WHERE owner_id=%s AND id IN %s
"""

SQL_BATCH_SELECT = """
SELECT id, owner_id, payload FROM {name}
    WHERE owner_id IS NULL AND topic=%s AND TIMESTAMPDIFF(SECOND, delivery_time, CURRENT_TIMESTAMP) >= 0
    ORDER BY priority DESC, id ASC
    LIMIT %s FOR UPDATE SKIP LOCKED;
"""
SQL_BATCH_UPDATE = "UPDATE {name} SET owner_id=%s WHERE id IN %s"

# Finish a message
SQL_ACK = "DELETE FROM {name} WHERE id=%s"

# Count messages in topic
SQL_GET_TOPIC_SIZE = "SELECT count(1) FROM {name} WHERE topic=%s AND owner_id IS NULL"

# Count messages in all topics
SQL_LIST_TOPICS = """
SELECT topic, count(*) FROM {name}
    WHERE owner_id IS NULL AND TIMESTAMPDIFF(SECOND, delivery_time, CURRENT_TIMESTAMP) >= 0
    GROUP BY topic
"""


class MySQLBackend(Backend):
    def __init__(self, connection, prefix: str):
        """
        :param connection: https://peps.python.org/pep-0249/#connection-objects
        """
        super().__init__()
        self.connection = connection
        self.prefix = prefix
        self.queue_table = f"{self.prefix}_queue"
        self.lock_table = f"{self.prefix}_lock"
        self.rate_limit_table = f"{self.prefix}_limits"
        self.owner_id = random.randint(0, 2**32 - 1)

    @property
    def max_payload_size(self) -> Optional[int]:
        return 2047

    @property
    def hash_size(self) -> int:
        return 16

    def create(self) -> None:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                SQL_CREATE.format(
                    name=self.queue_table,
                    size=self.max_payload_size,
                    hash_size=self.hash_size,
                )
            )
            cur.execute(SQL_CREATE_LOCKS.format(name=self.lock_table))
            cur.execute(
                SQL_CREATE_RATE_LIMITS.format(
                    name=self.rate_limit_table, key_size=self.hash_size
                )
            )
            self.connection.commit()

    def destroy(self) -> None:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_DROP.format(name=self.queue_table))
            cur.execute(SQL_DROP.format(name=self.lock_table))
            cur.execute(SQL_DROP.format(name=self.rate_limit_table))
            self.connection.commit()

    def batch_put(
        self,
        data: Collection[Tuple[bytes, int, Optional[bytes]]],
        priority: int,
        delay: int,
        failure_base_delay: int,
        visibility_timeout: int,
    ) -> int:
        for payload, topic, hsh in data:
            if len(payload) > self.max_payload_size:
                raise ValueError(
                    f"payload exceeds PAYLOAD_MAX_SIZE ({len(payload)} > {self.max_payload_size})"
                )
            if hsh is not None and len(hsh) != self.hash_size:
                raise ValueError(
                    f"hsh size is not HASH_SIZE ({len(hsh)} != {self.hash_size})"
                )

        with self.connection.cursor() as cur:
            self.connection.begin()
            rows = [
                (
                    payload,
                    topic,
                    hsh,
                    priority,
                    delay,
                    failure_base_delay,
                    visibility_timeout,
                )
                for payload, topic, hsh in data
            ]
            total = 0
            for row in rows:
                try:
                    cur.execute(
                        SQL_INSERT.format(name=self.queue_table),
                        args=row,
                    )
                    total += cur.rowcount
                except Exception as err:
                    if err.__class__.__name__ != "IntegrityError":
                        raise
            self.connection.commit()
            return total

    def release_stalled_messages(self) -> int:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_UPDATE.format(name=self.queue_table))
            rows = cur.rowcount
            self.connection.commit()
            return rows

    def batch_get(self, topic: int, size: int) -> List["Message"]:
        with self.connection.cursor() as cur:
            self.connection.begin()

            cur.execute(
                SQL_BATCH_SELECT.format(name=self.queue_table), args=(topic, size)
            )

            rows = cur.fetchall()
            if len(rows) == 0:
                self.connection.rollback()
                return []

            idxes = [x[0] for x in rows]
            cur.execute(
                SQL_BATCH_UPDATE.format(name=self.queue_table),
                args=(self.owner_id, idxes),
            )

            self.connection.commit()

        return [Message(x[2], x[0], self) for x in rows]

    def ack(self, task_id: int) -> None:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_ACK.format(name=self.queue_table), args=(task_id,))
            # TODO raise if it's already expired
            self.connection.commit()

    def batch_nack(self, task_ids: Collection[int]) -> None:
        if len(task_ids) == 0:
            return
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                SQL_BATCH_NACK_1.format(name=self.queue_table),
                args=(self.owner_id, list(task_ids)),
            )
            cur.execute(
                SQL_BATCH_NACK_2.format(name=self.queue_table),
                args=(self.owner_id, list(task_ids)),
            )
            # TODO raise if it's already expired
            self.connection.commit()

    def batch_touch(self, task_ids: Collection[int]) -> None:
        if len(task_ids) == 0:
            return
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                SQL_BATCH_TOUCH.format(name=self.queue_table),
                args=(self.owner_id, list(task_ids)),
            )
            # TODO raise if it's already expired
            self.connection.commit()

    def get_topic_size(self, topic: int) -> int:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_GET_TOPIC_SIZE.format(name=self.queue_table), args=(topic,))
            result = cur.fetchone()
            self.connection.commit()
            return result[0]

    def list_topics(self) -> List[Tuple[int, int]]:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_LIST_TOPICS.format(name=self.queue_table))
            rows = cur.fetchall()
            self.connection.commit()
        return rows

    def acquire_topic(
        self, topic_lock_visibility_timeout: int
    ) -> Optional["TopicLock"]:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(SQL_LIST_TOPICS.format(name=self.queue_table))
            topics = list(cur.fetchall())
            cur.execute(f"LOCK TABLE {self.lock_table} WRITE")

            topics.sort(key=lambda x: -x[1])
            new_lock = None
            for topic, count in topics:
                try:
                    cur.execute(
                        f"INSERT INTO {self.lock_table} (topic, owner_id, visibility_timeout) VALUES (%s, %s, %s)",
                        args=(topic, self.owner_id, topic_lock_visibility_timeout),
                    )
                except Exception:  # XXX
                    continue
                new_lock = topic
                break

            self.connection.commit()

            if new_lock is not None:
                return TopicLock(new_lock, self)
            return None

    def batch_release_topic(self, topics: Collection[int]) -> None:
        if len(topics) == 0:
            return
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                f"""
            DELETE FROM {self.lock_table} WHERE topic IN %s AND owner_id = %s
            """,
                args=(list(topics), self.owner_id),
            )
            self.connection.commit()

    def batch_touch_topic(self, topics: Collection[int]) -> None:
        if len(topics) == 0:
            return
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                f"""
            UPDATE {self.lock_table} SET acquire_time=CURRENT_TIMESTAMP
            WHERE topic IN %s AND owner_id = %s
            """,
                args=(list(topics), self.owner_id),
            )
            self.connection.commit()

    def release_stalled_topic_locks(self) -> int:
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                f"""
            DELETE FROM {self.lock_table} WHERE
            TIMESTAMPDIFF(SECOND, acquire_time, CURRENT_TIMESTAMP) > visibility_timeout
            """
            )
            rows = cur.rowcount
            self.connection.commit()
            return rows

    def rate_limit(self, key: bytes, max_events: int, interval_seconds: int) -> bool:
        if len(key) != self.hash_size:
            raise ValueError(
                f"rate limit key size is not HASH_SIZE ({len(key)} != {self.hash_size})"
            )
        with self.connection.cursor() as cur:
            self.connection.begin()
            cur.execute(
                f"DELETE FROM {self.rate_limit_table} "
                f"WHERE expire_time < CURRENT_TIMESTAMP"
            )
            cur.execute(
                f"SELECT COUNT(*) FROM {self.rate_limit_table} WHERE hash=%s",
                args=(key,),
            )
            (n,) = cur.fetchone()
            if n >= max_events:
                self.connection.commit()
                return False
            cur.execute(
                f"INSERT INTO {self.rate_limit_table} "
                f"SET hash=%s, expire_time=TIMESTAMPADD(SECOND, %s, CURRENT_TIMESTAMP)",
                args=(key, interval_seconds),
            )
            self.connection.commit()
            return True



from typing import Optional, Dict, List, Any, Generator

from squeal.backend.base import Message
from squeal.queue import Queue
from squeal.utils import get_logger, lm

logger = get_logger()


class BufferMessage(Message):
    FIELDS = Message.FIELDS + ["buffer"]

    def __init__(self, *args, buffer: "Buffer", **kwargs):
        super().__init__(*args, **kwargs)
        self.buffer = buffer

    @classmethod
    def from_message(cls, msg: Message, *args, **kwargs) -> "BufferMessage":
        kwargs.update({k: getattr(msg, k) for k in msg.FIELDS})
        return cls(*args, **kwargs)

    def ack(self):
        super().ack()
        self.buffer.ack(self.idx)

    def nack(self):
        super().nack()
        self.buffer.nack(self.idx)


class Buffer:
    def __init__(
        self,
        queue: Queue,
        extra_buffer_multiplier: int = 2,
        default_topic_quota: int = 1,
    ):
        self.queue = queue

        self.topic_buffer: Dict[int, List[Message]] = {}
        self.extra_buffer_multiplier = extra_buffer_multiplier
        self.default_topic_quota: int = default_topic_quota
        self.topic_quota: Dict[int, int] = {}
        self.topic_processing: Dict[int, int] = {}
        self.message_topic: Dict[int, int] = {}

        self.closed = False

    def close(self):
        if self.closed:
            raise RuntimeError
        self.closed = True
        self.queue.release_topics()
        self.queue.nack_all()

    def touch(self) -> None:
        if self.closed:
            raise RuntimeError
        self.queue.touch_all()
        self.queue.touch_topics()

    def _fill_buffer(self, idx: int) -> None:
        held = len(self.topic_buffer[idx])
        quota = self.topic_quota[idx]
        processing = self.topic_processing[idx]
        if held - processing >= quota:
            logger.debug(
                lm(
                    "Buffer._fill_buffer()",
                    {
                        "action": "none",
                        "idx": idx,
                        "n_held": held,
                        "n_quota": quota,
                        "n_processing": processing,
                    },
                )
            )
            return
        target = self.extra_buffer_multiplier * quota - held + processing
        msgs = self.queue.batch_get([(idx, target)])
        self.topic_buffer[idx].extend(msgs)
        logger.debug(
            lm(
                "Buffer._fill_buffer()",
                {
                    "action": "get",
                    "idx": idx,
                    "n_held": held,
                    "n_quota": quota,
                    "n_processing": processing,
                    "target": target,
                    "new_msgs": len(msgs),
                },
            )
        )

    def _acquire_topic(self) -> bool:
        topic = self.queue.acquire_topic()
        if topic is None:
            logger.debug(lm("Buffer._acquire_topic()", {"result": "failure"}))
            return False
        self.topic_quota[topic.idx] = self.default_topic_quota
        self.topic_processing[topic.idx] = 0
        self.topic_buffer[topic.idx] = []
        logger.debug(
            lm(
                "Buffer._acquire_topic()",
                {
                    "result": "success",
                    "topic_id": topic.idx,
                },
            )
        )
        self._fill_buffer(topic.idx)
        return True

    def _drop_topic(self, topic: int):
        if self.topic_processing[topic] > 0:
            raise RuntimeError

        logger.debug(
            lm(
                "Buffer._drop_topic()",
                {"topic_id": topic, "buffer_size": len(self.topic_buffer[topic])},
            )
        )

        for msg in self.topic_buffer[topic]:
            msg.nack()
        self.queue.release_topic(topic)

        del self.topic_buffer[topic]
        del self.topic_processing[topic]
        del self.topic_quota[topic]

    def _fill_buffers(self) -> None:
        # Check all the checked-out topics and make sure we have messages in the buffer
        # according to the topic's quota
        need_new_topic = True
        for topic_lock in self.queue.list_held_topics():
            topic = topic_lock.idx
            self._fill_buffer(topic)

            allowed = self.topic_quota[topic] - self.topic_processing[topic]
            if self.topic_buffer[topic] and allowed > 0:
                need_new_topic = False

            if self.topic_processing[topic] == 0 and not self.topic_buffer[topic]:
                self._drop_topic(topic)

        if need_new_topic:
            self._acquire_topic()

    def get(self) -> Optional[BufferMessage]:
        if self.closed:
            raise RuntimeError
        self._fill_buffers()
        out = None
        for topic_lock in self.queue.list_held_topics():
            topic = topic_lock.idx
            q = self.topic_quota[topic]
            p = self.topic_processing[topic]
            b = len(self.topic_buffer[topic])
            allowed = q - p
            logger.debug(
                lm(
                    "Buffer.get() topic",
                    {
                        "topic_id": topic,
                        "n_held": b,
                        "n_quota": q,
                        "n_processing": p,
                    },
                )
            )
            if allowed <= 0 or b == 0:
                continue

            msg = self.topic_buffer[topic].pop(-1)
            self.message_topic[msg.idx] = topic
            self.topic_processing[topic] += 1
            out = BufferMessage.from_message(msg, buffer=self)
            break
        logger.debug(
            lm(
                "Buffer.get() result",
                {
                    "success": ("true" if out is not None else "false"),
                },
            )
        )
        return out

    def ack(self, message_idx: int) -> None:
        if self.closed:
            raise RuntimeError
        topic = self.message_topic[message_idx]
        self.topic_processing[topic] -= 1

        logger.debug(
            lm(
                "Buffer.ack()",
                {
                    "topic_id": topic,
                    "n_processing": self.topic_processing[topic],
                },
            )
        )

    def nack(self, message_idx: int) -> None:
        if self.closed:
            raise RuntimeError
        topic = self.message_topic[message_idx]
        self.topic_processing[topic] -= 1

        logger.debug(
            lm(
                "Buffer.nack()",
                {
                    "topic_id": topic,
                    "n_processing": self.topic_processing[topic],
                },
            )
        )

    def __iter__(self) -> Generator[BufferMessage, Any, None]:
        while True:
            msg = self.get()
            if msg is None:
                break
            yield msg



from abc import ABC
from typing import List, Tuple, Optional, Collection


class Backend(ABC):
    def __init__(self, *args, **kwargs):
        pass

    @property
    def max_payload_size(self) -> Optional[int]:
        raise NotImplementedError

    @property
    def hash_size(self) -> int:
        raise NotImplementedError

    def create(self) -> None:
        raise NotImplementedError

    def destroy(self) -> None:
        raise NotImplementedError

    def batch_put(
        self,
        data: Collection[Tuple[bytes, int, Optional[bytes]]],
        priority: int,
        delay: int,
        failure_base_delay: int,
        visibility_timeout: int,
    ) -> int:
        raise NotImplementedError

    def release_stalled_messages(self) -> int:
        raise NotImplementedError

    def ack(self, task_id: int) -> None:
        raise NotImplementedError

    def batch_get(self, topic: int, size: int) -> List["Message"]:
        raise NotImplementedError

    def batch_nack(self, task_ids: Collection[int]) -> None:
        raise NotImplementedError

    def batch_touch(self, task_ids: Collection[int]) -> None:
        raise NotImplementedError

    def list_topics(self) -> List[Tuple[int, int]]:
        raise NotImplementedError

    def get_topic_size(self, topic: int) -> int:
        raise NotImplementedError

    def acquire_topic(
        self, topic_lock_visibility_timeout: int
    ) -> Optional["TopicLock"]:
        raise NotImplementedError

    def batch_release_topic(self, topics: Collection[int]) -> None:
        raise NotImplementedError

    def batch_touch_topic(self, topics: Collection[int]) -> None:
        raise NotImplementedError

    def release_stalled_topic_locks(self) -> int:
        raise NotImplementedError

    def rate_limit(self, key: bytes, max_events: int, interval_seconds: int) -> bool:
        raise NotImplementedError


class TopicLock:
    def __init__(self, idx: int, backend: Backend):
        self.idx = idx
        self.backend = backend
        self.released = False

    def __str__(self):
        return f"TopicLock({self.idx})"

    def release(self):
        if self.released:
            raise RuntimeError("Lock has already been released")
        self.backend.batch_release_topic([self.idx])
        self.released = True

    def touch(self):
        if self.released:
            raise RuntimeError("Lock has already been released")
        self.backend.batch_touch_topic([self.idx])


class Message:
    FIELDS = ["payload", "idx", "backend", "status"]

    def __init__(
        self, payload: bytes, idx: int, backend: Backend, status: Optional[bool] = None
    ):
        self.payload = payload
        self.idx = idx
        self.backend = backend
        self.status = status

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.status is None:
            self.nack()

    @property
    def released(self):
        return self.status is not None

    def ack(self):
        if self.released:
            raise RuntimeError("Message has already been relinquished")
        self.status = True
        self.backend.ack(self.idx)

    def nack(self):
        if self.released:
            raise RuntimeError("Message has already been relinquished")
        self.status = False
        self.backend.batch_nack([self.idx])

    def touch(self):
        if self.released:
            raise RuntimeError("Message has already been relinquished")
        self.backend.batch_touch([self.idx])

    def check(self) -> bool:
        """
        Check whether the message is still owned by this consumer.
        Use a local estimate based on when the message was acquired.
        """
        if self.released:
            return False

        raise NotImplementedError



from .queue import Queue
from .buffer import Buffer, BufferMessage
from .backend.base import Message
from .backend.mysql import MySQLBackend
from .utils import get_logger, enable_structured_logging, set_log_level



from collections import Counter
from typing import Type

import pytest

from squeal import Queue, Buffer
from .common import TemporaryMySQLBackend, TemporaryLocalBackend, TemporaryBackendMixin


@pytest.mark.parametrize(
    "backend_class", [TemporaryMySQLBackend, TemporaryLocalBackend]
)
class TestBuffer:
    def test_buffer_activity(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"a", 1, None)] * 100 + [(b"b", 2, None)] * 100,
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            buf = Buffer(Queue(bk))

            msg = buf.get()
            msg2 = buf.get()

            assert {msg.payload, msg2.payload} == {b"a", b"b"}

            msg3 = buf.get()
            assert msg3 is None

            msg.ack()
            msg4 = buf.get()
            assert msg4.payload == msg.payload  # same topic as the one we acked

            msg5 = buf.get()
            assert msg5 is None

    def test_buffer_works_through_large_queue(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                [(b"a", 1, None)] * 100 + [(b"b", 2, None)] * 100,
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            buf = Buffer(Queue(bk))

            seen = Counter()
            while True:
                msg = buf.get()
                if msg is None:
                    break
                seen[msg.payload] += 1
                msg.ack()

            assert dict(seen.items()) == {b"a": 100, b"b": 100}

    def test_buffer_finish_queue(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            buf = Buffer(Queue(bk))

            msg = buf.get()
            assert msg.payload == b"a"
            msg.ack()

            msg2 = buf.get()
            assert msg2 is None



import time
from typing import Type

import pytest

from .common import TemporaryMySQLBackend, TemporaryLocalBackend, TemporaryBackendMixin


@pytest.mark.parametrize(
    "backend_class", [TemporaryMySQLBackend, TemporaryLocalBackend]
)
class TestBackend:
    def test_create_destroy(self, backend_class: Type[TemporaryBackendMixin]):
        backend = backend_class()
        backend.create()
        backend.destroy()

    def test_release_stalled(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"test_release_stalled", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            x = bk.batch_get(topic=1, size=1)[0]
            assert b"test_release_stalled" == x.payload

            time.sleep(1)
            assert 1 == bk.release_stalled_messages()

            y = bk.batch_get(topic=1, size=1)[0]
            assert b"test_release_stalled" == y.payload

    def test_ack(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"test_ack", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            x = bk.batch_get(topic=1, size=1)[0]
            assert b"test_ack" == x.payload
            x.ack()

            time.sleep(1)
            bk.release_stalled_messages()

            assert 0 == len(bk.batch_get(topic=1, size=1))

    def test_nack(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"test_ack", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            x = bk.batch_get(topic=1, size=1)[0]
            assert b"test_ack" == x.payload

            assert 0 == len(bk.batch_get(topic=1, size=1))

            x.nack()

            z = bk.batch_get(topic=1, size=1)[0]
            assert z is not None

    def test_message_context_manager(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"test_ack", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            with bk.batch_get(topic=1, size=1)[0] as task:
                assert task is not None
                pass

            assert not task.status
            time.sleep(1)

            with bk.batch_get(topic=1, size=1)[0] as task:
                assert task is not None
                task.ack()
                pass

            assert task.status
            time.sleep(1)

            assert 0 == len(bk.batch_get(topic=1, size=1))

    def test_priority(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            bk.batch_put(
                [(b"b", 1, None)],
                priority=1,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            msg = bk.batch_get(topic=1, size=1)[0]
            assert b"b" == msg.payload

    def test_batch_get_empty(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            msgs = bk.batch_get(topic=1, size=2)
            assert 0 == len(msgs)

    def test_batch_get_less_than_full(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"b", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(topic=1, size=2)
            assert 1 == len(msgs)
            assert 0 == bk.get_topic_size(topic=1)

    def test_batch_get_full(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"b", 1, None)] * 2,
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(topic=1, size=2)
            assert 2 == len(msgs)
            assert 0 == bk.get_topic_size(topic=1)

    def test_batch_get_overfull(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                [(b"b", 1, None)] * 3,
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(topic=1, size=2)
            assert 2 == len(msgs)
            assert 1 == bk.get_topic_size(topic=1)

    def test_batch_put(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None), (b"b", 1, None), (b"c", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            assert 3 == bk.get_topic_size(topic=1)
            msgs = bk.batch_get(topic=1, size=3)
            assert 3 == len(msgs)
            assert 0 == bk.get_topic_size(topic=1)

    def test_uniqueness(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                data=[
                    (b"a", 1, b"0000000000000000"),
                    (b"b", 1, b"0000000000000001"),
                    (b"c", 1, b"0000000000000002"),
                ],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            assert bk.get_topic_size(1) == 3

    def test_uniqueness_doesnt_preclude_redoing_messages(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, b"0000000000000000")],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            assert bk.get_topic_size(1) == 1
            msg = bk.batch_get(1, 1)[0]
            msg.ack()

            assert bk.get_topic_size(1) == 0

            bk.batch_put(
                data=[(b"a", 1, b"0000000000000000")],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            assert bk.get_topic_size(1) == 1

    def test_uniqueness_violation(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            n_rows = bk.batch_put(
                data=[
                    (b"a", 1, b"0000000000000000"),
                    (b"b", 1, b"0000000000000001"),
                    (b"c", 1, b"0000000000000001"),
                ],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            assert n_rows == 2
            assert bk.get_topic_size(1) == 2

    def test_topic_acquire(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None), (b"b", 2, None), (b"c", 3, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            a = bk.acquire_topic(600)
            b = bk.acquire_topic(600)
            c = bk.acquire_topic(600)
            d = bk.acquire_topic(600)

            assert {a.idx, b.idx, c.idx} == {1, 2, 3}
            assert d is None

    def test_topic_timeout(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )

            a = bk.acquire_topic(1)
            assert a.idx == 1

            assert 0 == bk.release_stalled_topic_locks()

            b = bk.acquire_topic(1)
            assert b is None

            time.sleep(2)

            assert 1 == bk.release_stalled_topic_locks()
            c = bk.acquire_topic(1)
            assert c.idx == 1

    def test_touching_no_messages(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_touch([])

    def test_nacking_no_messages(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_nack([])

    def test_releasing_no_topics(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_release_topic([])

    def test_touching_no_topics(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            bk.batch_touch_topic([])

    def test_touching_non_list_collection(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(1, 1)
            msgs = {m.idx: m for m in msgs}
            bk.batch_touch(msgs.keys())

    def test_nacking_non_list_collection(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(1, 1)
            msgs = {m.idx: m for m in msgs}
            bk.batch_nack(msgs.keys())

    def test_releasing_non_list_collection(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(1, 1)
            msgs = {m.idx: m for m in msgs}
            bk.batch_release_topic(msgs.keys())

    def test_touching_topics_non_list_collection(
        self, backend_class: Type[TemporaryBackendMixin]
    ):
        with backend_class() as bk:
            bk.batch_put(
                data=[(b"a", 1, None)],
                priority=0,
                delay=0,
                failure_base_delay=0,
                visibility_timeout=0,
            )
            msgs = bk.batch_get(1, 1)
            msgs = {m.idx: m for m in msgs}
            bk.batch_touch_topic(msgs.keys())

    def test_rate_limiting(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            key = b"0" * bk.hash_size
            for _ in range(10):
                assert bk.rate_limit(key, max_events=10, interval_seconds=60)
            for _ in range(10):
                assert not bk.rate_limit(key, max_events=10, interval_seconds=60)

    def test_rate_limiting_expires(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            key = b"0" * bk.hash_size
            for _ in range(2):
                assert bk.rate_limit(key, max_events=2, interval_seconds=1)
            for _ in range(2):
                assert not bk.rate_limit(key, max_events=2, interval_seconds=1)

            time.sleep(2)

            for _ in range(2):
                assert bk.rate_limit(key, max_events=2, interval_seconds=1)
            for _ in range(2):
                assert not bk.rate_limit(key, max_events=2, interval_seconds=1)

    def test_rate_limiting_bad_key(self, backend_class: Type[TemporaryBackendMixin]):
        with backend_class() as bk:
            key = b"0" * bk.hash_size + b"0"
            with pytest.raises(ValueError):
                assert bk.rate_limit(key, max_events=10, interval_seconds=60)



import math
from typing import Tuple, List, Dict, Iterable, Optional, Collection
from squeal.backend.base import Backend, Message, TopicLock


class Queue:
    """
    FIFO(-ish) queue backed by a SQL table
    """

    def __init__(
        self,
        backend: Backend,
        new_message_delay: int = 0,
        failure_base_delay: int = 1,
        visibility_timeout: int = 60,
        topic_lock_visibility_timeout: int = 60 * 15,
        rate_limit_per_hour: Optional[float] = None,
        auto_create: bool = True,
    ):
        self.backend = backend
        self.new_message_delay = new_message_delay
        self.failure_base_delay = failure_base_delay
        self.visibility_timeout = visibility_timeout
        self.topic_lock_visibility_timeout = topic_lock_visibility_timeout
        self.rate_limit_per_hour = rate_limit_per_hour

        self._held_messages: Dict[int, "Message"] = {}
        self._held_topics: Dict[int, "TopicLock"] = {}

        if self.new_message_delay < 0:
            raise RuntimeError("Delay must be non-negative")
        if self.failure_base_delay < 0:
            raise RuntimeError("Delay must be non-negative")
        if self.visibility_timeout <= 0:
            raise RuntimeError("Visibility timeout must be positive")

        if auto_create:
            self.create()

    def create(self) -> None:
        self.backend.create()

    def destroy(self) -> None:
        self.backend.destroy()

    @property
    def _rate_limit_interval_seconds(self) -> Optional[int]:
        if self.rate_limit_per_hour is None:
            return None
        return math.ceil(60 * 60 / self.rate_limit_per_hour)

    def put(
        self, item: bytes, topic: int, hsh: Optional[bytes] = None, priority: int = 0
    ) -> int:
        return self.batch_put(items=[(item, topic, hsh)], priority=priority)

    def batch_put(
        self, items: Collection[Tuple[bytes, int, Optional[bytes]]], priority: int = 0
    ) -> int:
        interval = self._rate_limit_interval_seconds
        if interval is not None:
            approved_items = []
            for item, topic, hsh in items:
                if hsh is None or self.backend.rate_limit(
                    hsh, max_events=1, interval_seconds=interval
                ):
                    approved_items.append((item, topic, hsh))
            items = approved_items

        return self.backend.batch_put(
            items,
            priority,
            self.new_message_delay,
            self.failure_base_delay,
            self.visibility_timeout,
        )

    def get(self, topic: int) -> Optional["Message"]:
        msgs = self.batch_get([(topic, 1)])
        return msgs[0] if msgs else None

    def batch_get(self, topics: Iterable[Tuple[int, int]]) -> List["Message"]:
        out = []
        self.backend.release_stalled_messages()
        for topic, size in topics:
            if size <= 0:
                continue
            topic_msgs = self.backend.batch_get(topic, size)
            out.extend(topic_msgs)
        for msg in out:
            self._held_messages[msg.idx] = msg
        return out

    def _prune_held_messages(self):
        # XXX
        released = [msg.idx for msg in self._held_messages.values() if msg.released]
        for idx in released:
            del self._held_messages[idx]

    def touch_all(self) -> None:
        # XXX
        self._prune_held_messages()
        self.backend.batch_touch(self._held_messages.keys())

    def nack_all(self) -> None:
        # XXX
        self._prune_held_messages()
        self.backend.batch_nack(self._held_messages.keys())
        self._held_messages.clear()

    def list_topics(self) -> List[Tuple[int, int]]:
        return self.backend.list_topics()

    def get_topic_size(self, topic: int) -> int:
        return self.backend.get_topic_size(topic)

    def _prune_held_topics(self):
        # XXX
        released = [i for i, t in self._held_topics.items() if t.released]
        for i in released:
            del self._held_topics[i]

    def acquire_topic(self) -> Optional["TopicLock"]:
        self.backend.release_stalled_topic_locks()
        topic = self.backend.acquire_topic(
            topic_lock_visibility_timeout=self.topic_lock_visibility_timeout
        )
        if topic is None:
            return None
        self._held_topics[topic.idx] = topic
        return topic

    def release_topics(self) -> None:
        self._prune_held_topics()
        self.backend.batch_release_topic(self._held_topics.keys())
        self._held_topics = {}

    def release_topic(self, topic: int) -> None:
        self._prune_held_topics()
        if topic not in self._held_topics:
            return
        self._held_topics[topic].release()
        self._prune_held_topics()

    def touch_topics(self) -> None:
        self._prune_held_topics()
        self.backend.batch_touch_topic(self._held_topics.keys())

    def list_held_topics(self) -> List["TopicLock"]:
        self._prune_held_topics()
        return list(self._held_topics.values())


__all__ = ["Queue"]
